'''
Created on 6 dec. 2024

@author: tleduc

Copyright 2020-2024 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from pandas import concat
from pytz import timezone
from pandas import Timestamp, date_range
from datetime import datetime, timedelta

from t4gpd.commons.IllegalArgumentTypeException import IllegalArgumentTypeException
from t4gpd.commons.LatLonLib import LatLonLib
from t4gpd.commons.sun.PySolarSunLib import PySolarSunLib
from t4gpd.commons.sun.SoleneSunLib import SoleneSunLib

import matplotlib.pyplot as plt


class SunModel(object):
    '''
    classdocs
    '''
    class Pvlib(object):
        def __init__(self, gdf, altitude):
            from pvlib.location import Location

            lat, lon = LatLonLib.fromGeoDataFrameToLatLon(gdf)
            self.location = Location(lat, lon, altitude)

        def clearsky_irradiances(self, dts, model="ineichen"):
            return self.location.get_clearsky(dts, model)

        def positions(self, dts):
            return self.location.get_solarposition(dts)

        def positions_clearsky_irradiances(self, dts, model="ineichen"):
            positions = self.location.get_solarposition(dts)
            irradiances = self.location.get_clearsky(dts, model)
            result = concat([positions, irradiances], axis=1)
            # result.reset_index(names="timestamp", inplace=True)
            return result

    class CircularEarthOrbit(object):
        def __init__(self, gdf, altitude):
            self.lat, _ = LatLonLib.fromGeoDataFrameToLatLon(gdf)

        @staticmethod
        def __solar_declination(day_of_year):
            # http://www.heliodon.net/downloads/Beckers_2010_Helio_007_fr.pdf
            # Spencer formula
            from numpy import cos, pi, sin
            g = ((2 * pi) * (day_of_year - 1)) / 365
            return (0.006918 +
                    -0.399912 * cos(g) + 0.070257 * sin(g)
                    - 0.006758 * cos(2 * g) + 0.000907 * sin(2 * g)
                    - 0.002697 * cos(3 * g) + 0.00148 * sin(3 * g))

        @staticmethod
        def __hour_angle(dt):
            # Observing the sun from earth, the solar hour angle
            # is an expression of time, expressed in angular
            # measurement, usually degrees, from solar noon. At
            # solar noon the hour angle is 0.0 degree. The time
            # before solar noon is expressed as negative degrees.
            # 24hrs correspond to 360 degrees, 15 degrees per hour.
            from numpy import deg2rad
            return deg2rad(((dt.hour + (dt.minute / 60)) - 12) * 15)

        @staticmethod
        def __solar_angles(latitude, declination, hour_angle):
            from numpy import arcsin, arctan2, cos, deg2rad, pi, rad2deg, sin
            from t4gpd.commons.AngleLib import AngleLib

            latitude = deg2rad(latitude)

            # ELEVATION
            elevation = arcsin(
                sin(latitude) * sin(declination)
                + cos(latitude) * cos(declination) * cos(hour_angle)
            )

            # AZIMUTH
            azimuth = arctan2(
                (-cos(declination) * sin(hour_angle)) / cos(elevation),
                (sin(elevation) * sin(latitude) - sin(declination)) /
                (cos(elevation) * cos(latitude))
            )
            azimuth = AngleLib.southCCW2northCW(azimuth, degree=False)

            return rad2deg([elevation, azimuth])

        def clearsky_irradiances(self, dts):
            from t4gpd.energy.PerrinDeBrichambaut import PerrinDeBrichambaut
            df = self.positions(dts)

            raise NotImplementedError("Must be implemented!")

        def positions(self, dts):
            from pandas import DataFrame
            df = DataFrame({"elevation": dts, "azimuth": dts}, index=dts)
            df = DataFrame(
                columns=["day_of_year", "declination",
                         "hour_angle", "elev_azim", "elevation", "azimuth"],
                index=dts)
            df = df.assign(
                day_of_year=df.index.day_of_year,
                declination=lambda row: self.__solar_declination(
                    row.day_of_year),
                hour_angle=lambda row: self.__hour_angle(
                    row.index),
            )
            df.elev_azim = df.apply(lambda row: self.__solar_angles(
                self.lat, row.declination, row.hour_angle), axis=1)
            df.elevation = df.elev_azim.apply(lambda ea: ea[0])
            df.azimuth = df.elev_azim.apply(lambda ea: ea[1])
            df.drop(columns=["elev_azim"], inplace=True)
            return df

        def positions_clearsky_irradiances(self, dts):
            return self.clearsky_irradiances(dts)

    def __init__(self, gdf=LatLonLib.NANTES, altitude=0, model="pvlib"):
        '''
        Constructor
        '''
        model_class = SunModel.model_switch(model)
        self.model = model_class(gdf, altitude)
        self.lat, self.lon = LatLonLib.fromGeoDataFrameToLatLon(gdf)
        self.altitude = altitude

    @ staticmethod
    def __check_DatetimeIndex(dts):
        from pandas import DatetimeIndex
        if not isinstance(dts, DatetimeIndex):
            raise IllegalArgumentTypeException(
                dts, "pandas.DatetimeIndex")

    @ staticmethod
    def model_switch(modelName):
        modelName = modelName.lower()
        if ("pvlib" == modelName):
            return SunModel.Pvlib
        elif ("circular" == modelName):
            return SunModel.CircularEarthOrbit
        raise IllegalArgumentTypeException(
            modelName, "model as 'pvlib' or 'circular'")

    def sun_rise_set(self, dts, tz="UTC"):
        from pandas import DataFrame
        from suntimes.suntimes import SunTimes

        SunModel.__check_DatetimeIndex(dts)
        st = SunTimes(self.lon, self.lat, altitude=self.altitude)
        df = DataFrame({"sunrise": dts, "sunset": dts}, index=dts)
        df.sunrise = df.sunrise.apply(lambda dt: st.riseutc(dt))
        df.sunrise = df.sunrise.dt.tz_localize("UTC")
        df.sunset = df.sunset.apply(lambda dt: st.setutc(dt))
        df.sunset = df.sunset.dt.tz_localize("UTC")
        if ("UTC" != tz):
            df.sunrise = df.sunrise.dt.tz_convert(tz)
            df.sunset = df.sunset.dt.tz_convert(tz)
        df = df.assign(daylength=df.sunset-df.sunrise)
        # df.daylength = df.daylength.apply(lambda v: v.seconds//60)
        return df

    def clearsky_irradiances(self, dts):
        SunModel.__check_DatetimeIndex(dts)
        return self.model.clearsky_irradiances(dts)

    def positions(self, dts):
        SunModel.__check_DatetimeIndex(dts)
        return self.model.positions(dts)

    def positions_clearsky_irradiances(self, dts):
        SunModel.__check_DatetimeIndex(dts)
        return self.model.positions_clearsky_irradiances(dts)

    def getDayLengthInMinutes(self, dt):
        return self.model.getDayLengthInMinutes(dt)

    def getFractionalYear(self, dt):
        return self.model.getFractionalYear(dt)

    def getRadiationDirection(self, dt):
        return self.model.getRadiationDirection(dt)

    def getSolarAnglesInDegrees(self, dt):
        return self.model.getSolarAnglesInDegrees(dt)

    def getSolarAnglesInRadians(self, dt):
        return self.model.getSolarAnglesInRadians(dt)

    def getSolarDeclination(self, dayOfYear):
        return self.model.getSolarDeclination(dayOfYear)

    def getSunrise(self, dt):
        return self.model.getSunrise(dt)

    def getSunset(self, dt):
        return self.model.getSunset(dt)

    @staticmethod
    def polarPlot(models=["circular", "pvlib"], year=2023):
        from numpy import deg2rad
        from pandas import DatetimeIndex, Timestamp

        solstices_equinox = DatetimeIndex([
            Timestamp(f"{year}-03-21"),
            Timestamp(f"{year}-06-21"),
            Timestamp(f"{year}-12-21"),
        ])

        ncols = len(models)
        fig, axes = plt.subplots(figsize=(1.2*8.26, 0.5*8.26),
                                 nrows=1, ncols=ncols, squeeze=False,
                                 subplot_kw={"projection": "polar"})
        for nc, model in enumerate(models):
            sun_model = SunModel(model=model)
            sun_rise_set = sun_model.sun_rise_set(solstices_equinox, tz="UTC")

            ax = axes[0, nc]
            # NORTH IS ON TOP
            ax.set_theta_zero_location("N")
            # CLOCKWISE
            ax.set_theta_direction(-1)
            # LABEL POSITIONS
            # ax.set_rlabel_position(0)
            ax.set_title(model)
            ax.set_ylim(0, 90)

            for month, color in [(3, "green"), (6, "red"), (12, "blue")]:
                day = Timestamp(f"{year}-{month}-21")
                sunrise = sun_rise_set.loc[day, "sunrise"]
                sunset = sun_rise_set.loc[day, "sunset"]
                dts = date_range(sunrise, sunset, freq="10min")
                df = sun_model.positions(dts)
                ax.plot(deg2rad(df.azimuth), 90-df.elevation, color=color)

            for hour in range(6, 19, 2):
                start = Timestamp(f"{year}-01-01 {hour:02d}:00")
                end = Timestamp(f"{year}-12-31 {hour:02d}:00")
                dts = date_range(start, end, freq="5    D")
                df = sun_model.positions(dts)
                ax.plot(deg2rad(df.azimuth), 90-df.elevation, color="brown",
                        linewidth=0.75, linestyle="solid")

            ax.set_rgrids(
                range(0, 90, 10),
                labels=[f"{a}°" for a in range(90, 0, -10)])
            ax.set_thetagrids(
                range(0, 360, 15),
                labels=[f"{a}°" for a in range(0, 360, 15)])
        plt.show()
        plt.close(fig)

    def plotDayLengths(self):
        year = datetime.now().year
        daysInYear = range(1, self.model.nDaysPerYear(year) + 1, 10)
        dayLengths = []

        for _dayOfYear in daysInYear:
            _dt = datetime(year, 1, 1, 12, tzinfo=timezone.utc) + \
                timedelta(days=_dayOfYear)
            _dayLengthInMin = self.model.getDayLengthInMinutes(_dt)
            dayLengths.append(_dayLengthInMin)

        if (isinstance(self.model, SoleneSunLib)):
            _title = "Solene model"
        elif (isinstance(self.model, PySolarSunLib)):
            _title = "PySolar implementation"

        plt.title("%s: days lengths - Year %d\nLatitude: %.1f$^o$, Longitude: %.1f$^o$" % (
            _title, year, self.lat, self.lon))
        plt.xlabel("Day of year")
        plt.ylabel("Day length (min.)")
        plt.grid()
        plt.plot(daysInYear, dayLengths)
        plt.show()

    def plotSolarDeclination(self):
        _year = datetime.now().year
        _the1stOfJan = datetime(_year, 1, 1, 12, tzinfo=timezone.utc)

        days = range(1, self.model.nDaysPerYear(_year) + 1)

        declinations = [
            self.model.getSolarDeclination(
                _the1stOfJan + timedelta(days=dayOfYear))
            for dayOfYear in days]

        if (isinstance(self.model, SoleneSunLib)):
            plt.title(
                "Solar declination by Solene (Spencer, 1971) - year = %d" % _year)
        elif (isinstance(self.model, PySolarSunLib)):
            plt.title("Solar declination by PySolar - year = %d" % _year)

        plt.xlabel("Day of year")
        plt.ylabel("Solar declination")
        plt.grid()
        plt.plot(days, declinations)
        plt.show()

    def plotSunAltiAtNoon(self):
        _year = datetime.now().year
        _the1stOfJan = datetime(_year, 1, 1, 12, tzinfo=timezone.utc)

        daysInYear = range(1, self.model.nDaysPerYear(_year) + 1, 10)
        sunAltiAtNoon = []

        for _dayOfYear in daysInYear:
            _dt = _the1stOfJan + timedelta(days=_dayOfYear)
            _alti, _ = self.model.getSolarAnglesInDegrees(_dt)
            sunAltiAtNoon.append(_alti)

        if (isinstance(self.model, SoleneSunLib)):
            _title = "Solene model"
        elif (isinstance(self.model, PySolarSunLib)):
            _title = "PySolar implementation"

        plt.title("%s: days lengths - Year %d\nLatitude: %.1f$^o$, Longitude: %.1f$^o$" % (
            _title, _year, self.lat, self.lon))
        plt.xlabel("Day of year")
        plt.ylabel("Solar altitude at noon (in degrees)")
        plt.grid()
        plt.plot(daysInYear, sunAltiAtNoon)
        plt.show()

    def plotSunriseSunset(self):
        year = datetime.now().year
        daysInYear = range(1, self.model.nDaysPerYear(year) + 1, 10)
        sunrises, sunsets = [], []

        for _dayOfYear in daysInYear:
            _dt = datetime(year, 1, 1, 12, tzinfo=timezone.utc) + \
                timedelta(days=_dayOfYear)

            _sunrise = self.model.getTimeSpentSinceMidnight(
                self.model.getSunrise(_dt))
            _sunset = self.model.getTimeSpentSinceMidnight(
                self.model.getSunset(_dt))

            sunrises.append(_sunrise)
            sunsets.append(_sunset)

        if (isinstance(self.model, SoleneSunLib)):
            _title = "Solene model"
        elif (isinstance(self.model, PySolarSunLib)):
            _title = "PySolar implementation"

        plt.title("%s: sunrise/sunset - Year %d\nLatitude: %.1f$^o$, Longitude: %.1f$^o$" % (
            _title, year, self.lat, self.lon))
        plt.xlabel("Day of year")
        plt.ylabel("UTC Time (hr.)")
        plt.grid()
        plt.plot(daysInYear, sunrises, daysInYear, sunsets)
        plt.show()

    def plotSolarPanorama(self):
        year = datetime.now().year
        _, basemap = plt.subplots(figsize=(1 * 8.26, 1 * 8.26))

        if (isinstance(self.model, SoleneSunLib)):
            _title = "Solene model"
        elif (isinstance(self.model, PySolarSunLib)):
            _title = "PySolar implementation"
        plt.title("%s: solar panorama - year %d\nLatitude: %.1f$^o$, Longitude: %.1f$^o$" % (
            _title, year, self.lat, self.lon))
        plt.xlabel("Azim. [$^0$]")
        plt.ylabel("Alti. [$^0$]")

        altis, azims = {}, {}
        hrsAltis, hrsAzims = {}, {}

        for hour in range(0, 24):
            hrsAltis[hour], hrsAzims[hour] = [], []

        for month in range(6, 13, 3):
            altis[month], azims[month] = [], []
            t0 = datetime(year, month, 21, tzinfo=timezone.utc)
            _sunrise = self.model.getSunrise(t0)
            _sunset = self.model.getSunset(t0)

            for hour in range(0, 24):
                for minute in range(0, 60, 10):
                    _dt = t0 + timedelta(hours=hour) + \
                        timedelta(minutes=minute)
                    if True or (_sunrise < _dt < _sunset):
                        _alti, _azim = self.model.getNativeSolarAnglesInDegrees(
                            _dt)
                        if (0 <= _alti):
                            altis[month].append(_alti)
                            azims[month].append(_azim)
                        if (0 == minute):
                            hrsAltis[hour].append(_alti)
                            hrsAzims[hour].append(_azim)

            basemap.plot(azims[month], altis[month],
                         label=_dt.strftime("%b %d"))

        for hour in range(5, 20, 1):
            if (0 == (hour % 2)):
                basemap.plot(hrsAzims[hour], hrsAltis[hour], "k-.")
                basemap.text((3 * hrsAzims[hour][0] + hrsAzims[hour][-1]) / 4,
                             (3 * hrsAltis[hour][0] + hrsAltis[hour][-1]) / 4,
                             f"{hour}:00", fontsize=9).set_color("black")

        for lbl, azim in [("East", 90), ("South", 180), ("West", 270)]:
            basemap.axvline(x=azim, linestyle=":", color="black", linewidth=1)
            basemap.text(azim, 70, lbl, fontsize=9).set_color("black")
        basemap.set_ylim(0, 90)

        basemap.legend()
        plt.show()


"""
tz = "Europe/Paris"
tz = "UTC"
dt0 = Timestamp("2024-06-21 08:00", tz=tz)
dt1 = Timestamp("2024-06-21 18:00", tz=tz)
dts = date_range(dt0, dt1, freq="1h")

# dt0 = Timestamp("2024-03-21 08:00", tz=tz)
# dt1 = Timestamp("2024-09-21 08:00", tz=tz)
# dts = date_range(dt0, dt1, freq="1ME")

# model = "pvlib"
# df = SunModel(model=model).sun_rise_set(dts, tz=tz)

model = "pvlib"
df1 = SunModel(model=model).positions(dts)

model = "circular"
df2 = SunModel(model=model).positions(dts)

print(
    concat([df1[["elevation", "azimuth"]], df2[["elevation", "azimuth"]]], axis=1)
)

model = "pvlib"
df3 = SunModel(model=model).positions_clearsky_irradiances(dts)

SunModel.polarPlot()
"""



'''
Created on 27 mar. 2024

@author: tleduc

Copyright 2020-2024 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from datetime import date, datetime, timedelta, timezone
from geopandas import GeoDataFrame
from numpy import asarray, cos, dot, heaviside, matmul, ndarray, repeat, stack, zeros
from t4gpd.commons.AngleLib import AngleLib
from t4gpd.commons.ArrayCoding import ArrayCoding
from t4gpd.commons.GeoProcess import GeoProcess
from t4gpd.commons.IllegalArgumentTypeException import IllegalArgumentTypeException
from t4gpd.commons.sun.SunLib import SunLib
from t4gpd.energy.PerrinDeBrichambaut import PerrinDeBrichambaut


class SkyMapRadiationBalance2(GeoProcess):
    '''
    classdocs
    '''

    def __init__(self, skymaps, dt, meteo=None, svfFieldname="svf",
                 anglesFieldname="angles", model="pysolar", encode=False):
        '''
        Constructor
        '''
        if not isinstance(skymaps, GeoDataFrame):
            raise IllegalArgumentTypeException(skymaps, "GeoDataFrame")
        self.skymaps = skymaps

        if not svfFieldname in skymaps:
            raise Exception(f"{svfFieldname} is not a relevant field name!")
        self.svf = svfFieldname

        if not anglesFieldname in skymaps:
            raise Exception(f"{anglesFieldname} is not a relevant field name!")
        self.angles = anglesFieldname

        sunModel = SunLib(gdf=skymaps, model=model)
        if isinstance(dt, datetime):
            day = date(*dt.timetuple()[0:3])
            dt0 = sunModel.getSunrise(day) + timedelta(hours=1)
            dt1 = dt
        elif isinstance(dt, date):
            dt0 = sunModel.getSunrise(dt) + timedelta(hours=1)
            dt1 = sunModel.getSunset(dt)
        elif (isinstance(dt, (list, ndarray, tuple)) and (2 == len(dt)) and
              isinstance(dt[0], datetime) and isinstance(dt[1], datetime)):
            dt0, dt1 = dt
            if (date(*dt0.timetuple()[0:3]) != date(*dt1.timetuple()[0:3])):
                raise IllegalArgumentTypeException(
                    dt, "Pair of datetimes must occur on the same day!")
            if (dt0 > dt1):
                raise IllegalArgumentTypeException(
                    dt, "Pair of datetimes must be ordered!")
        else:
            raise IllegalArgumentTypeException(
                dt, "single date or datetime; or pair of datetimes")

        self.dts = [
            datetime(*dt0.timetuple()[0:3], h, tzinfo=timezone.utc)
            for h in range(dt0.hour, dt1.hour+1)
        ]

        if meteo is None:
            self.irrad = SkyMapRadiationBalance2.__theoretical_irrad(
                sunModel, self.dts)
        else:
            if "timestamp" in meteo:
                self.irrad = SkyMapRadiationBalance2.__irrad_from_daily_meteo(
                    meteo, self.dts)
            else:
                self.irrad = SkyMapRadiationBalance2.__irrad_from_monthly_meteo(
                    meteo, self.dts)
        self.encode = encode

    @staticmethod
    def __irrad_from_daily_meteo(meteo, dts):
        cols = ["hour", "sun_beam_dir", "solar_alti",
                "solar_azim", "GHI", "DNI", "DHI",]
        _meteo = meteo.loc[meteo[
            (dts[0] <= meteo.timestamp) & (meteo.timestamp <= dts[-1])
        ].index, cols]
        _meteo.set_index("hour", drop=True, inplace=True)
        return _meteo.to_dict(orient="index")

    @staticmethod
    def __irrad_from_monthly_meteo(meteo, dts):
        cols = ["hour", "sun_beam_dir", "solar_alti",
                "solar_azim", "GHI", "DNI", "DHI"]
        month = dts[0].month
        h0, h1 = dts[0].hour, dts[-1].hour

        _meteo = meteo.loc[meteo[
            (meteo.month == month) & (meteo.hour.isin(range(h0, h1+1)))
        ].index, cols]
        _meteo.set_index("hour", drop=True, inplace=True)
        return _meteo.to_dict(orient="index")

    @staticmethod
    def __theoretical_irrad(sunModel, dts, skyType=PerrinDeBrichambaut.STANDARD_SKY):
        result = {}
        for dt in dts:
            alti, azim = sunModel.getSolarAnglesInDegrees(dt)
            sbdir = sunModel.getRadiationDirection(dt)
            dni, dhi, ghi = 0, 0, 0
            if (0.0 < alti):
                dni = PerrinDeBrichambaut.directNormalIrradiance(
                    AngleLib.toRadians(alti), skyType=skyType)
                dhi = PerrinDeBrichambaut.diffuseSolarIrradiance(
                    AngleLib.toRadians(alti), skyType)
                # GHI = DNI x cos(theta) + DHI
                ghi = dni * cos(AngleLib.toRadians(90-alti)) + dhi

            result[dt.hour] = {
                "sun_beam_dir": sbdir,
                "solar_alti": alti,
                "solar_azim": azim,
                "GHI": ghi,
                "DNI": dni,
                "DHI": dhi
            }
        return result

    def run(self):
        hours = sorted(self.irrad.keys())
        nangles = len(self.skymaps.loc[self.skymaps.index.min(), self.angles])
        nhours, nsensors = len(hours), len(self.skymaps)

        # ========== DHI PRE-PROCESSING ==========
        dhi = asarray([self.irrad[h]["DHI"] for h in hours])
        dhi = dhi.sum()

        # ========== DNI PRE-PROCESSING ==========
        sunAzim = asarray([self.irrad[h]["solar_azim"] for h in hours])
        offset = 360 / nangles
        sunAzim = ((sunAzim + offset/2) / offset).astype(int) % 64
        delta = zeros([nangles, nhours])
        for c, l in enumerate(sunAzim):
            delta[l, c] = 1

        sunAlti = asarray([self.irrad[h]["solar_alti"] for h in hours])
        sunAlti = sunAlti.reshape(1, -1)
        sunAlti = repeat(sunAlti, nsensors, axis=0)
        sunAlti[sunAlti < 0] = 0

        M = stack(self.skymaps[self.angles].to_numpy())
        H = heaviside(sunAlti - matmul(M, delta), 0)
        H = H.astype(int)

        dniMagn = asarray([self.irrad[h]["DNI"] for h in hours])
        dni = asarray([self.irrad[h]["sun_beam_dir"] for h in hours])
        normal = (0, 0, 1)
        dni = dot(dni, normal)
        dni[dni < 0] = 0
        dni = dniMagn * dni

        dni = matmul(H, dni)

        # PROCESSING
        hrs_in_sun = H.sum(axis=1)
        result = self.skymaps.copy(deep=True)
        result["in_the_sun"] = H.tolist()
        result["hrs_in_sun"] = hrs_in_sun
        result["hrs_shade"] = nhours-hrs_in_sun
        result["sw_diffuse"] = dhi * result[self.svf]
        result["sw_direct"] = dni
        return result


"""
import matplotlib.pyplot as plt
from pandas import read_csv, to_datetime
from t4gpd.commons.ArrayCoding import ArrayCoding
from t4gpd.demos.GeoDataFrameDemos import GeoDataFrameDemos
from t4gpd.morph.STGrid import STGrid
from t4gpd.morph.STSkyMap25D import STSkyMap25D
from t4gpd.morph.geoProcesses.STGeoProcess import STGeoProcess

buildings = GeoDataFrameDemos.ensaNantesBuildings()
sensors = STGrid(buildings, dx=80, dy=None, indoor=False,
                 intoPoint=True, withDist=False).run()  # < 0.1 sec.
skymaps = STSkyMap25D(buildings, sensors, nRays=64, rayLength=100,
                      elevationFieldname="HAUTEUR", h0=0.0, size=9.0,
                      withIndices=True, withAngles=True, encode=False).run()  # < 1 sec.

day = date(2022, 6, 13)  # Focus on the hottest day
# day = date(2022, 12, 22)  # Focus on the coldest day
# irrad = SkyMapRadiationBalance2(skymaps, day, meteo=None).run()

meteo = read_csv(
    "/home/tleduc/prj/spatial_distribution_of_solar_radiation/data/meteorology_extreme.csv")
meteo.timestamp = meteo.timestamp.apply(lambda dt: to_datetime(dt))

# meteo = read_csv("/home/tleduc/prj/spatial_distribution_of_solar_radiation/data/meteorology.csv")
meteo.sun_beam_dir = meteo.sun_beam_dir.apply(lambda dt: ArrayCoding.decode(dt))

irrad2 = SkyMapRadiationBalance2(skymaps, day, meteo=meteo).run()
print(irrad2[["gid", "svf", "in_the_sun", "hrs_in_sun", "hrs_shade", "sw_direct", "sw_diffuse"]])

fig, ax = plt.subplots(figsize=(1.0*8.26, 1.0*8.26))
buildings.plot(ax=ax)
irrad2.set_geometry("viewpoint").plot(ax=ax, column="sw_direct", legend=True)
irrad2.set_geometry("viewpoint").apply(lambda x: ax.annotate(
	text=x.gid, xy=x.geometry.centroid.coords[0],
	color="black", size=12, ha="center"), axis=1)
fig.tight_layout()
ax.axis("off")
plt.show()
plt.close(fig)
"""



'''
Created on 25 oct 2023

@author: tleduc

Copyright 2020-2023 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from dijkstar.graph import Graph
from geopandas import GeoDataFrame, overlay
from networkx import DiGraph
from shapely import LineString, union_all
from t4gpd.commons.GeomLib import GeomLib
from t4gpd.commons.IllegalArgumentTypeException import IllegalArgumentTypeException
from t4gpd.commons.graph.UrbanGraph import UrbanGraph
from t4gpd.commons.graph.UrbanGraphVertex import UrbanGraphVertex
from warnings import warn


class UrbanGraphFactory(object):
    '''
    classdocs
    '''

    @staticmethod
    def create(roads, method=None):
        if not isinstance(roads, GeoDataFrame):
            raise IllegalArgumentTypeException(roads, "GeoDataFrame")

        geometryCollection = union_all(roads.geometry)
        listOfBipoints = GeomLib.toListOfBipointsAsLineStrings(
            geometryCollection)
        listOfBipoints = [GeomLib.removeZCoordinate(
            bp) for bp in listOfBipoints]

        if method is None:
            ciVertices, icVertices, edges = UrbanGraphFactory.__M1_build_graph(
                listOfBipoints)
            return UrbanGraph(ciVertices=ciVertices, icVertices=icVertices,
                              edges=edges, crs=roads.crs)

        elif ("networkx" == method):
            # USEFUL FOR CALCULATING MULTI-CENTRALITIES, MST
            gdfOfBipoints = GeoDataFrame(
                {"gid": range(len(listOfBipoints)), "geometry": listOfBipoints}, crs=roads.crs)
            ciVertices, icVertices, nx_graph = UrbanGraphFactory.__M2_build_graph(
                listOfBipoints)
            return UrbanGraph(ciVertices=ciVertices, icVertices=icVertices,
                              nx_graph=nx_graph, gdfOfBipoints=gdfOfBipoints,
                              crs=roads.crs)

        elif ("dijkstar" == method):
            ciVertices, icVertices, edges, dk_graph = UrbanGraphFactory.__M3_build_graph(
                listOfBipoints)
            return UrbanGraph(ciVertices=ciVertices, icVertices=icVertices,
                              edges=edges, dk_graph=dk_graph, crs=roads.crs)

        raise IllegalArgumentTypeException(
            method, "None, 'networkx' or 'dijkstar'")

    @ staticmethod
    def __M1_build_graph(listOfBipoints):
        ciVertices = dict()
        icVertices = dict()
        edges = dict()

        for linestring in listOfBipoints:
            UrbanGraphFactory.__M1_add_linestring(
                ciVertices, icVertices, edges, linestring)

        return ciVertices, icVertices, edges

    @ staticmethod
    def __M1_add_linestring(ciVertices, icVertices, edges, linestring):
        coords = linestring.coords
        if (2 <= len(coords)):
            first = UrbanGraphVertex.add_vertex(
                ciVertices, icVertices, coords[0])
            last = UrbanGraphVertex.add_vertex(
                ciVertices, icVertices, coords[-1])

            if (first in edges) and (first == last):
                warn("This is a multigraph (loop)!")
                # Handle multigraph (two vertices may be connected by more than one edge)
                newEdges1 = GeomLib.cutsLineStringByCurvilinearDistance(
                    linestring, linestring.length / 3.0)
                newEdges2 = GeomLib.cutsLineStringByCurvilinearDistance(
                    linestring, 2.0 * linestring.length / 3.0)
                newEdges = newEdges1 + newEdges2[:-1]
                for _edge in newEdges:
                    UrbanGraphFactory.__M1_add_linestring(
                        ciVertices, icVertices, edges, _edge)

            elif (first in edges) and (last in edges[first]):
                warn("This is a multigraph!")
                # Handle multigraph (two vertices may be connected by more than one edge)
                newEdges = GeomLib.cutsLineStringByCurvilinearDistance(
                    linestring, linestring.length / 2.0)
                for _edge in newEdges:
                    UrbanGraphFactory.__M1_add_linestring(
                        ciVertices, icVertices, edges, _edge)

            else:
                UrbanGraphFactory.__M1_add_pair_or_vertices(
                    edges, first, last, linestring)
                UrbanGraphFactory.__M1_add_pair_or_vertices(
                    edges, last, first, LineString(reversed(coords)))

    @staticmethod
    def __M1_add_pair_or_vertices(edges, first, last, linestring):
        linestringLength = linestring.length

        edgeItem = {"geometry": linestring, "length": linestringLength}
        if first in edges:
            if last in edges[first]:
                raise Exception("Unreachable instruction!")
            else:
                edges[first][last] = edgeItem
        else:
            edges[first] = dict({last: edgeItem})
        pass

    @staticmethod
    def __M2_build_graph(listOfBipoints):
        ciVertices = dict()
        icVertices = dict()
        nx_graph = DiGraph()

        multigraph = 0
        for linestring in listOfBipoints:
            multigraph += UrbanGraphFactory.__M2_add_linestring(
                ciVertices, icVertices, nx_graph, linestring)

        if 0 < multigraph:
            warn(f"Handle {multigraph} multigraphs!")

        return ciVertices, icVertices, nx_graph

    @staticmethod
    def __M2_add_linestring(ciVertices, icVertices, nx_graph, linestring):
        multigraph = 0
        coords = linestring.coords
        if (2 <= len(coords)):
            first = UrbanGraphVertex.add_vertex(
                ciVertices, icVertices, coords[0])
            last = UrbanGraphVertex.add_vertex(
                ciVertices, icVertices, coords[-1])

            if (first in nx_graph) and (last in nx_graph[first]):
                multigraph = 1
                # Handle multigraph (two vertices may be connected by more than one edge)
                newEdges = GeomLib.cutsLineStringByCurvilinearDistance(
                    linestring, linestring.length / 2.0)
                for _edge in newEdges:
                    UrbanGraphFactory.__M2_add_linestring(
                        ciVertices, icVertices, nx_graph, _edge)
            else:
                UrbanGraphFactory.__M2_add_pair_or_vertices(
                    nx_graph, first, last, linestring)
                UrbanGraphFactory.__M2_add_pair_or_vertices(
                    nx_graph, last, first, LineString(reversed(coords)))
        return multigraph

    @staticmethod
    def __M2_add_pair_or_vertices(nx_graph, first, last, linestring):
        nx_graph.add_edge(first, last, weight=linestring.length, object={
            "geometry": linestring, "weight": linestring.length})

    @staticmethod
    def __M3_build_graph(listOfBipoints):
        ciVertices = dict()
        icVertices = dict()
        edges = dict()
        dk_graph = Graph()

        for linestring in listOfBipoints:
            UrbanGraphFactory.__M3_add_linestring(
                ciVertices, icVertices, edges, dk_graph, linestring)

        return ciVertices, icVertices, edges, dk_graph

    @staticmethod
    def __M3_add_linestring(ciVertices, icVertices, edges, dk_graph,  linestring):
        coords = linestring.coords
        if (2 <= len(coords)):
            first = UrbanGraphVertex.add_vertex(
                ciVertices, icVertices, coords[0])
            last = UrbanGraphVertex.add_vertex(
                ciVertices, icVertices, coords[-1])

            if (first in edges) and (first == last):
                warn("This is a multigraph (loop)!")
                # Handle multigraph (two vertices may be connected by more than one edge)
                newEdges1 = GeomLib.cutsLineStringByCurvilinearDistance(
                    linestring, linestring.length / 3.0)
                newEdges2 = GeomLib.cutsLineStringByCurvilinearDistance(
                    linestring, 2.0 * linestring.length / 3.0)
                newEdges = newEdges1 + newEdges2[:-1]
                for _edge in newEdges:
                    UrbanGraphFactory.__M3_add_linestring(
                        ciVertices, icVertices, edges, dk_graph, _edge)

            elif (first in edges) and (last in edges[first]):
                warn("This is a multigraph!")
                # Handle multigraph (two vertices may be connected by more than one edge)
                newEdges = GeomLib.cutsLineStringByCurvilinearDistance(
                    linestring, linestring.length / 2.0)
                for _edge in newEdges:
                    UrbanGraphFactory.__M3_add_linestring(
                        ciVertices, icVertices, edges, dk_graph, _edge)

            else:
                UrbanGraphFactory.__M3_add_pair_or_vertices(
                    edges, dk_graph, first, last, linestring)
                UrbanGraphFactory.__M3_add_pair_or_vertices(
                    edges, dk_graph, last, first, LineString(reversed(coords)))

    @staticmethod
    def __M3_add_pair_or_vertices(edges, dk_graph, first, last, linestring):
        linestringLength = linestring.length

        dk_graph.add_edge(first, last, (linestringLength, linestring.wkt))

        edgeItem = {"geometry": linestring, "length": linestringLength}
        if first in edges:
            if last in edges[first]:
                raise Exception("Unreachable instruction!")
            else:
                edges[first][last] = edgeItem
        else:
            edges[first] = dict({last: edgeItem})



'''
Created on 23 aug. 2024

@author: tleduc

Copyright 2020-2024 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from pandas import DatetimeIndex, to_datetime
from geopandas import GeoDataFrame
from pvlib.location import Location
from t4gpd.commons.GeoProcess import GeoProcess
from t4gpd.commons.IllegalArgumentTypeException import IllegalArgumentTypeException
from t4gpd.commons.LatLonLib import LatLonLib


class STPvlibIrradiances(GeoProcess):
    '''
    classdocs
    '''

    def __init__(self, gdf, dts, altitude=0, model="ineichen"):
        '''
        Constructor
        '''
        if not isinstance(gdf, GeoDataFrame):
            raise IllegalArgumentTypeException(gdf, "GeoDataFrame")

        self.dts = to_datetime(dts) if not isinstance(
            dts, DatetimeIndex) else dts
        lat, lon = LatLonLib.fromGeoDataFrameToLatLon(gdf)
        self.location = Location(lat, lon, altitude)

        self.model = model

    def run(self):
        df = self.location.get_clearsky(self.dts, model=self.model)
        # df = df.assign(timestamp=df.index)
        df.reset_index(names="timestamp", inplace=True)
        return df

    @staticmethod
    def test():
        import matplotlib.pyplot as plt
        from datetime import date, datetime, timedelta, timezone
        from locale import LC_ALL, setlocale
        from pandas import date_range
        from t4gpd.commons.sun.SunLib import SunLib

        sunModel = SunLib(LatLonLib.NANTES, model="pysolar")
        day = date(2023, 6, 21)
        sunrise, sunset = sunModel.getSunrise(day), sunModel.getSunset(day)
        dts = date_range(start=sunrise, end=sunset,
                         freq=f"60min", inclusive="neither",
                         tz=timezone.utc)

        # dt = datetime(*sunrise.timetuple()[:-3], tzinfo=timezone.utc)
        # dts = []
        # while (dt + timedelta(minutes=60) < sunset):
        #     dt += timedelta(minutes=60)
        #     dts.append(dt)

        df = STPvlibIrradiances(LatLonLib.NANTES, dts).run()
        df = df.assign(time=df.timestamp.apply(lambda dt: dt.time()))

        # PLOTTING
        fig, ax = plt.subplots(figsize=(8.26, 8.26))
        setlocale(LC_ALL, "en_US.utf8")
        ax.set_title(f"Irradiation, {day.strftime('%B %d')}", fontsize=20)
        ax1 = df.plot("time", "dni", ax=ax, grid=True, label="DNI")
        df.plot("time", "dhi", ax=ax, grid=True, label="DHI")
        df.plot("time", "ghi", ax=ax, grid=True, label="GHI")
        ax1.set_xlabel("Timestamp (UTC)", fontdict={"fontsize": 16})
        ax1.set_ylabel("[W.m$^{-2}$]", fontdict={'fontsize': 16})
        fig.tight_layout()
        plt.show()

        return df


# STPvlibIrradiances.test()



'''
Created on 10 nov. 2023

@author: tleduc

Copyright 2020-2024 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from numpy import cos, linspace, pi, sin, stack
from shapely import MultiLineString
from shapely.affinity import translate
from t4gpd.commons.GeomLib import GeomLib
from t4gpd.commons.GeomLib3D import GeomLib3D


class PanopticRaysLib(object):
    '''
    classdocs
    '''
    """
    NORTH, WEST, SOUTH, EAST = 1, 2, 3, 4
    NE = pi / 4
    NW = (3 * pi) / 4
    SW = (5 * pi) / 4
    SE = (7 * pi) / 4

    @staticmethod
    def __fromRadiansToCompassCode(angle):
        if (PanopticRaysLib.NE <= angle < PanopticRaysLib.NW):
            return PanopticRaysLib.NORTH
        if (PanopticRaysLib.NW <= angle < PanopticRaysLib.SW):
            return PanopticRaysLib.WEST
        if (PanopticRaysLib.SW <= angle < PanopticRaysLib.SE):
            return PanopticRaysLib.SOUTH
        return PanopticRaysLib.EAST

    @staticmethod
    def __fromCompassCodeToOrientation(geometry):
        code = geometry.coords[0][2]
        if (PanopticRaysLib.NORTH == code):
            return "north"
        if (PanopticRaysLib.WEST == code):
            return "west"
        if (PanopticRaysLib.SOUTH == code):
            return "south"
        return "east"

    """
    @staticmethod
    def __get2DMultiLineString(rayLength, nRays):
        angles = linspace(0, 2.0 * pi, nRays, endpoint=False)
        shootingDirs = rayLength * stack([cos(angles), sin(angles)], axis=1)
        return MultiLineString(
            [([(0, 0), xy]) for xy in shootingDirs])
        """
        shootingDirs = hstack([
            rayLength * stack([cos(angles), sin(angles)], axis=1),
            vectorize(PanopticRaysLib.__fromRadiansToCompassCode)(
                angles).reshape(-1, 1)
        ])
        return MultiLineString(
            [([(0, 0, xyz[2]), xyz]) for xyz in shootingDirs])
        """

    @staticmethod
    def get2DGeoDataFrame(sensors, rayLength=100.0, nRays=64):
        rays2D = sensors.copy(deep=True)
        rays2D["viewpoint"] = rays2D.loc[:, "geometry"]
        rays2D.geometry = rays2D.geometry.apply(
            lambda geom: geom.centroid)

        if (rayLength in sensors) and (nRays in sensors):
            rays2D.geometry = rays2D.apply(
                lambda row: translate(
                    PanopticRaysLib.__get2DMultiLineString(
                        row[rayLength], row[nRays]),
                    xoff=row.geometry.x,
                    yoff=row.geometry.y),
                axis=1)
        elif rayLength in sensors:
            rays2D.geometry = rays2D.apply(
                lambda row: translate(
                    PanopticRaysLib.__get2DMultiLineString(
                        row[rayLength], nRays),
                    xoff=row.geometry.x,
                    yoff=row.geometry.y),
                axis=1)
        elif nRays in sensors:
            rays2D.geometry = rays2D.apply(
                lambda row: translate(
                    PanopticRaysLib.__get2DMultiLineString(
                        rayLength, row[nRays]),
                    xoff=row.geometry.x,
                    yoff=row.geometry.y),
                axis=1)
        else:
            shootingDirs = PanopticRaysLib.__get2DMultiLineString(
                rayLength, nRays)
            rays2D.geometry = rays2D.geometry.apply(
                lambda geom: translate(shootingDirs, xoff=geom.x, yoff=geom.y))

        rays2D = rays2D.explode(index_parts=True)
        """
        rays2D["compass"] = rays2D.geometry.apply(
            lambda geom: PanopticRaysLib.__fromCompassCodeToOrientation(geom))
        rays2D.geometry = rays2D.geometry.apply(
            lambda geom: GeomLib.removeZCoordinate(geom))
        """
        rays2D = rays2D.reset_index(names=["__VPT_ID__", "__RAY_ID__"])
        return rays2D

    @staticmethod
    def get25DGeoDataFrame(sensors, rayLength=100.0, nRays=64, h0=0.0):
        rays25D = PanopticRaysLib.get2DGeoDataFrame(sensors, rayLength, nRays)
        rays25D.geometry = rays25D.apply(
            lambda row: GeomLib.forceZCoordinateToZ0(
                row.geometry, z0=GeomLib3D.centroid(
                    row.viewpoint).z if row.viewpoint.has_z else h0), axis=1)
        rays25D.viewpoint = rays25D.viewpoint.apply(
            lambda vp: GeomLib.forceZCoordinateToZ0(
                vp, z0=GeomLib3D.centroid(vp).z if vp.has_z else h0))
        return rays25D



'''
Created on 19 nov. 2024

@author: tleduc

Copyright 2020-2024 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from geopandas import GeoDataFrame
from pandas import concat
from t4gpd.commons.GeoProcess import GeoProcess
from t4gpd.commons.IllegalArgumentTypeException import IllegalArgumentTypeException
from t4gpd.morph.geoProcesses.STGeoProcess import STGeoProcess


class STSatelliteLOSAnalysis(GeoProcess):
    '''
    classdocs
    '''

    def __init__(self, lblsat):
        '''
        Constructor
        '''
        if not isinstance(lblsat, GeoDataFrame):
            raise IllegalArgumentTypeException(
                lblsat, "lblsat GeoDataFrame")

        self.lblsat = lblsat.query("label == 'LOS'", inplace=False)

    def __lin_reg(self, lblsat):
        from t4gpd.morph.geoProcesses.LinearRegressionIndices import LinearRegressionIndices

        op = LinearRegressionIndices(with_geom=True)
        linregsat = STGeoProcess(op, lblsat).run()
        return linregsat

    def __mbc(self, chull):
        from t4gpd.morph.geoProcesses.CircularityIndices import CircularityIndices

        op = CircularityIndices(with_geom=True)
        mbcsat = STGeoProcess(op, chull).run()
        return mbcsat

    def __mabe(self, chull):
        from t4gpd.morph.geoProcesses.EllipticityIndices import EllipticityIndices

        op = EllipticityIndices(threshold=None, debug=False, with_geom=True)
        mabesat = STGeoProcess(op, chull).run()
        return mabesat

    def __mabeaxis(self, lblsat):
        from t4gpd.morph.geoProcesses.EllipticityAxisIndices import EllipticityAxisIndices

        op = EllipticityAxisIndices(
            threshold=None, debug=False, with_geom=True)
        mabeaxissat = STGeoProcess(op, lblsat).run()
        return mabeaxissat

    def __mabr(self, chull):
        from t4gpd.morph.geoProcesses.RectangularityIndices import RectangularityIndices

        op = RectangularityIndices(with_geom=True)
        mabrsat = STGeoProcess(op, chull).run()
        return mabrsat

    def run(self):
        # GROUP SATELLITES BY GID AND BY CONSTELLATION
        lblsat = self.lblsat.dissolve(
            by=["gid", "timeUTC", "constellation"],
            aggfunc={"satellite": list}, as_index=False)

        linreg = self.__lin_reg(lblsat)

        # LET'S CONVERT XPOINTS INTO THE CORRESPONDING CONVEX HULL
        chull = lblsat.assign(geometry=lambda geom: geom.convex_hull)
        mbc = self.__mbc(chull)
        mabr = self.__mabr(chull)
        mabe = self.__mabe(chull)
        mabeaxis = self.__mabeaxis(lblsat)

        # CONCAT ALL LAYERS USING INDEX
        lblsat = concat([
            lblsat,
            linreg[["slope", "intercept", "score", "mae", "mse"]],
            mabe[["flattening", "a_elli_def", "p_elli_def"]],
            mabeaxis[["ell_mae", "ell_mse"]],
            mabr[["stretching", "a_rect_def", "p_rect_def"]],
            mbc[["gravelius", "jaggedness", "miller", "morton", "a_circ_def"]]
        ], axis=1)

        return lblsat, chull, linreg, mabe, mabeaxis, mabr, mbc

    @staticmethod
    def test():
        import matplotlib.pyplot as plt
        from datetime import timezone
        from geopandas import read_file
        from matplotlib.colors import ListedColormap
        from t4gpd.resilientgaia.STECEF2AERSatelliteReader import STECEF2AERSatelliteReader
        from t4gpd.resilientgaia.STSatelliteLabelling import STSatelliteLabelling
        from t4gpd.resilientgaia.STSatelliteOnSiteProjection import STSatelliteOnSiteProjection
        from t4gpd.skymap.STSkyMap25D import STSkyMap25D

        # LOAD DATASETS
        rootdir = "/home/tleduc/prj/anr_resilientgaia_2024-2028/dev/data"
        ifile = f"{rootdir}/20240917_ReSilientGAIA.gpkg"
        buildings = read_file(ifile, layer="buildings")

        ifile = f"{rootdir}/20240917_parcours_timestamp.csv"
        aersat = STECEF2AERSatelliteReader(
            ifile, timestampFieldName="timeUTC").run()
        aersat["gid"] = range(len(aersat))
        aersat = aersat.set_index("timeUTC").tz_convert(
            timezone.utc).reset_index()

        gids = [959, 1615, 1718]
        # gids = [959]

        aersat = aersat.query(f"gid in {gids}", inplace=False)

        # PROCESS DATASETS
        projectionName, size = "Stereographic", 4.0
        skymaps = STSkyMap25D(buildings, aersat, nRays=64, rayLength=100.0,
                              elevationFieldname="HAUTEUR", h0=0.0, size=size, epsilon=1e-2,
                              projectionName=projectionName, withIndices=True, withAngles=False,
                              encode=True, threshold=1e-6).run()
        prjsat = STSatelliteOnSiteProjection(aersat, gid="gid", timestamp="timeUTC",
                                             proj=projectionName, size=size).run()
        prjsat2 = STSatelliteLabelling(skymaps, prjsat, "gid").run()

        lblsat, chull, linreg, mabe, mabeaxis, mabr, mbc = STSatelliteLOSAnalysis(
            prjsat2).run()

        # MAPPING
        my_cmap = ListedColormap(["green", "red"])

        for gid in gids:
            minx, miny, maxx, maxy = skymaps.query(
                f"gid == {gid}").buffer(1.0).total_bounds
            fig, axes = plt.subplots(
                figsize=(1.75 * 8.26, 1.35 * 8.26), nrows=2, ncols=3, squeeze=False)

            _skymaps = skymaps.query(f"gid == {gid}", inplace=False)
            svf = _skymaps.svf.squeeze()
            w_mean = _skymaps.w_mean.squeeze()
            h_over_w = _skymaps.h_over_w.squeeze()

            title = f"Position {gid}\n(svf={svf:.2f}, w_mean={w_mean:.1f} m, h/w={h_over_w:.3f})"
            fig.suptitle(title, fontsize=20)
            for nc, constellation in enumerate(["Galileo", "GLONASS", "GPS"]):
                querystring = f"(constellation == '{constellation}') and (gid == {gid})"
                _prjsat2 = prjsat2.query(querystring, inplace=False)
                _chull = chull.query(querystring, inplace=False)
                _mabe = mabe.query(querystring, inplace=False)
                _mabeaxis = mabeaxis.query(querystring, inplace=False)
                _mabr = mabr.query(querystring, inplace=False)
                _mbc = mbc.query(querystring, inplace=False)
                _linreg = linreg.query(querystring, inplace=False)

                # =====
                ax = axes[0, nc]
                ax.set_title(constellation, fontsize=16)
                skymaps.plot(ax=ax, color="lightgrey")
                aersat.plot(ax=ax, color="black", marker="P")
                _prjsat2.plot(ax=ax, column="label", cmap=my_cmap, marker="o")

                _chull.boundary.plot(ax=ax, color="brown",
                                     linestyle="dashed", linewidth=1)
                _mabe.boundary.plot(ax=ax, color="blue",
                                    linestyle="dashed", linewidth=1)
                _mabeaxis.plot(ax=ax, color="blue",
                               linestyle="dashed", linewidth=1)
                _mabeaxis.apply(lambda x: ax.annotate(
                    text=f"flattening={x.flattening:.2f}\nMAE={x.ell_mae:.2f}\nMSE={x.ell_mse:.2f}",
                    xy=[minx, miny], color="black", size=12, ha="left", va="bottom"), axis=1)
                ax.axis("off")
                ax.axis([minx, maxx, miny, maxy])

                # =====
                ax = axes[1, nc]

                skymaps.plot(ax=ax, color="lightgrey")
                aersat.plot(ax=ax, color="black", marker="P")
                _prjsat2.plot(ax=ax, column="label", cmap=my_cmap, marker="o")

                _mbc.boundary.plot(ax=ax, color="brown",
                                   linestyle="dashed", linewidth=1)
                _mabr.boundary.plot(ax=ax, color="blue",
                                    linestyle="dashed", linewidth=1)
                _linreg.plot(ax=ax, color="blue",
                             linestyle="dashed", linewidth=1)
                _linreg.apply(lambda x: ax.annotate(
                    text=f"score={x.score:.2f}\nMAE={x.mae:.2f}\nMSE={x.mse:.2f}",
                    xy=[minx, miny], color="black", size=12, ha="left", va="bottom"), axis=1)
                _mabr.apply(lambda x: ax.annotate(
                    text=f"stretching={x.stretching:.2f}\na_rect_def={x.a_rect_def:.2f}\np_rect_def={x.p_rect_def:.2f}",
                    xy=[maxx, miny], color="black", size=12, ha="right", va="bottom"), axis=1)
                _mbc.apply(lambda x: ax.annotate(
                    text=f"jaggedness={x.jaggedness:.2f}\nmiller={x.miller:.2f}\na_circ_def={x.a_circ_def:.2f}",
                    xy=[maxx, maxy], color="black", size=12, ha="right", va="top"), axis=1)
                ax.axis("off")
                ax.axis([minx, maxx, miny, maxy])

            plt.savefig(f"satmap_{gid}.pdf", bbox_inches="tight")
            # plt.show()
            plt.close(fig)

        return lblsat, chull, linreg, mabe, mabeaxis, mabr, mbc


# lblsat, chull, linreg, mabe, mabeaxis, mabr, mbc = STSatelliteLOSAnalysis.test()

"""
import matplotlib.pyplot as plt
from datetime import timezone
from geopandas import read_file
from matplotlib.colors import ListedColormap
from t4gpd.resilientgaia.STECEF2AERSatelliteReader import STECEF2AERSatelliteReader
from t4gpd.resilientgaia.STSatelliteLabelling import STSatelliteLabelling
from t4gpd.resilientgaia.STSatelliteOnSiteProjection import STSatelliteOnSiteProjection
from t4gpd.skymap.STSkyMap25D import STSkyMap25D

# LOAD DATASETS
ifile = "/home/tleduc/prj/anr_resilientgaia_2024-2028/dev/data/20240917_ReSilientGAIA.gpkg"
buildings = read_file(ifile, layer="buildings")

ifile = "/home/tleduc/prj/anr_resilientgaia_2024-2028/dev/data/20240917_parcours_timestamp.csv"
aersat = STECEF2AERSatelliteReader(ifile, timestampFieldName="timeUTC").run()
aersat["gid"] = range(len(aersat))
aersat = aersat.set_index("timeUTC").tz_convert(timezone.utc).reset_index()
gids = [959, 1615, 1718]
# gids = [959]
aersat = aersat.loc[aersat[aersat.gid.isin(gids)].index]

# PROCESS DATASETS
projectionName, size = "Stereographic", 4.0
skymaps = STSkyMap25D(buildings, aersat, nRays=64, rayLength=100.0,
                      elevationFieldname="HAUTEUR", h0=0.0, size=size, epsilon=1e-2,
                      projectionName=projectionName, withIndices=True, withAngles=False,
                      encode=True, threshold=1e-6).run()
prjsat = STSatelliteOnSiteProjection(aersat, gid="gid", timestamp="timeUTC",
                                     proj=projectionName, size=size).run()
prjsat2 = STSatelliteLabelling(skymaps, prjsat, "gid").run()

lblsat, chull, linreg, mabe, mabeaxis, mabr, mbc = STSatelliteLOSAnalysis(prjsat2).run()
"""



'''
Created on 10 nov. 2023

@author: tleduc

Copyright 2020-2024 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from geopandas import GeoDataFrame, overlay, sjoin_nearest
from numpy import asarray, full, mean, min, zeros
from pandas import concat
from shapely import equals, LineString, Point
from shapely.ops import nearest_points
from t4gpd.commons.GeoDataFrameLib import GeoDataFrameLib
from t4gpd.commons.GeomLib import GeomLib
from t4gpd.commons.GeomLib3D import GeomLib3D
from t4gpd.commons.SVFLib import SVFLib
from t4gpd.commons.raycasting.PanopticRaysLib import PanopticRaysLib


class RayCasting25DLib(object):
    '''
    classdocs
    '''
    @staticmethod
    def get25DPanopticRaysGeoDataFrame(sensors, rayLength=100.0, nRays=64, h0=0.0):
        return PanopticRaysLib.get25DGeoDataFrame(sensors, rayLength, nRays, h0)

    @staticmethod
    def __build_LineString(pA, pB, threshold):
        if ((pA.is_empty) or (pB.is_empty) or (pA.distance(pB) <= threshold)):
            return LineString([])
        return LineString([pA, pB])

    @staticmethod
    def __keepTheLargestSolidAngle(sensor, rays, threshold):
        sensor = GeomLib3D.centroid(sensor)
        if isinstance(rays, (Point, LineString)):
            _, remotePoint = nearest_points(sensor, rays)
            # Debug 10.11.2023: nearest_points returns 2D points
            remotePoint = GeomLib.forceZCoordinateToZ0(
                remotePoint, z0=rays.coords[0][2])
            return RayCasting25DLib.__build_LineString(sensor, remotePoint, threshold)

        maxSolidAngle, maxMask = -float("inf"), None
        for mask in rays.geoms:
            h, w = mask.coords[0][2], sensor.distance(mask)
            if (0 == w):
                return LineString([])
            currSolidAngle = h / w
            if (currSolidAngle > maxSolidAngle):
                maxSolidAngle, maxMask = currSolidAngle, mask
        _, remotePoint = nearest_points(sensor, maxMask)
        # Debug 10.11.2023: nearest_points returns 2D points
        remotePoint = GeomLib.forceZCoordinateToZ0(
            remotePoint, z0=maxMask.coords[0][2])

        return RayCasting25DLib.__build_LineString(sensor, remotePoint, threshold)

    @staticmethod
    def __from25DRaysToRayLengths(nRays, ray_ids, mls):
        _raylens = asarray(
            [ray.length for ray in GeomLib.toListOfLineStrings(mls)])

        if (nRays == len(_raylens)):
            return _raylens

        raylens = zeros(nRays)
        for i, ray_id in enumerate(ray_ids):
            raylens[ray_id] = _raylens[i]
        return raylens

    @staticmethod
    def __from25DRaysToRayAlts(nRays, ray_ids, mls, default_height):
        _rayalts = asarray(
            [ray.coords[-1][2] for ray in GeomLib.toListOfLineStrings(mls)])

        if (nRays == len(_rayalts)):
            return _rayalts

        # The altitude of the building adjacent to the viewpoint should be assigned here.
        # rayalts = zeros(nRays)
        rayalts = full(nRays, default_height)
        for i, ray_id in enumerate(ray_ids):
            rayalts[ray_id] = _rayalts[i]
        return rayalts

    @staticmethod
    def __from25DRaysToRayDeltaAlts(nRays, ray_ids, mls, default_height):
        _rays = GeomLib.toListOfLineStrings(mls)
        _rayDeltaAlts = asarray(
            [max(ray.coords[-1][2]-ray.coords[0][2], 0) for ray in _rays])

        if (nRays == len(_rayDeltaAlts)):
            return _rayDeltaAlts

        # The altitude of the building adjacent to the viewpoint should be assigned here.
        # rayDeltaAlts = zeros(nRays)
        alt0 = min([ray.coords[0][2] for ray in _rays], initial=0)
        rayDeltaAlts = full(nRays, default_height-alt0)
        for i, ray_id in enumerate(ray_ids):
            rayDeltaAlts[ray_id] = _rayDeltaAlts[i]
        return rayDeltaAlts

    @staticmethod
    def multipleRayCast25D(viewpoints, buildings, rays, nRays, elevationFieldName,
                           withIndices, h0=0.0, threshold=1e-9):
        if not GeoDataFrameLib.shareTheSameCrs(buildings, rays):
            raise Exception(
                "Illegal argument: buildings and rays must share shames CRS!")
        # rays.to_csv("/tmp/1.csv", index=False) # DEBUG

        rays.geometry = rays.geometry.apply(lambda geom: GeomLib.forceZCoordinateToZ0(
            geom, z0=geom.coords[0][2] if geom.has_z else h0))

        smapRaysField = overlay(rays, buildings[["geometry", elevationFieldName]],
                                how="intersection", keep_geom_type=True)

        if (0 == len(smapRaysField)):
            smapRaysField = rays.set_index("__VPT_ID__", drop=True).set_index(
                "__RAY_ID__", append=True, drop=False)
            smapRaysField["__HAUTEUR_VP__"] = None
        else:
            # Debug 25.03.2024
            smapRaysField.geometry = smapRaysField.apply(
                lambda row: GeomLib.forceZCoordinateToZ0(
                    row.geometry, row[elevationFieldName]),
                axis=1)
            # smapRaysField.to_csv("/tmp/2.csv") # DEBUG

            # Debug 06.05.2024
            # smapRaysField = smapRaysField.dissolve(
            #     by=["__VPT_ID__", "__RAY_ID__"], as_index=False, aggfunc={
            #     "gid": "first", "viewpoint": "first"})
            smapRaysField = smapRaysField.dissolve(
                by=["__VPT_ID__", "__RAY_ID__"], as_index=False, aggfunc="first")
            # smapRaysField.to_csv("/tmp/3.csv") # DEBUG
            smapRaysField.geometry = smapRaysField.apply(lambda row: RayCasting25DLib.__keepTheLargestSolidAngle(
                row.viewpoint, row.geometry, threshold), axis=1)
            # smapRaysField.to_csv("/tmp/31.csv") # DEBUG

            left = rays.set_index(["__VPT_ID__", "__RAY_ID__"], drop=False)
            left.drop(columns=["__VPT_ID__"], inplace=True)
            # left.to_csv("/tmp/32.csv") # DEBUG
            right = smapRaysField.set_index(
                ["__VPT_ID__", "__RAY_ID__"]).geometry.to_frame()
            # right.to_csv("/tmp/33.csv") # DEBUG
            smapRaysField = left.merge(
                right, how="left", left_index=True, right_index=True)
            # smapRaysField.to_csv("/tmp/34.csv") # DEBUG

            smapRaysField["geometry"] = smapRaysField.apply(
                lambda row: row.geometry_x if row.geometry_y is None else row.geometry_y, axis=1)
            smapRaysField.drop(
                columns=["geometry_x", "geometry_y"], inplace=True)
            # Debug 20.08.2024
            smapRaysField = GeoDataFrame(smapRaysField).set_crs(
                buildings.crs, allow_override=True)

            # Retrieve the heights of buildings near problematic sensors
            # Debug 25.03.2024
            indices1 = smapRaysField[smapRaysField.geometry.apply(
                lambda geom: geom.is_empty)].index
            ids1 = smapRaysField.loc[indices1, "gid"].unique()
            smapRaysField.drop(index=indices1, inplace=True)

            viewpoints2 = viewpoints.loc[viewpoints.gid.isin(ids1), [
                "gid", "geometry"]]
            viewpoints2 = sjoin_nearest(
                viewpoints2, buildings[["geometry", elevationFieldName]])
            viewpoints2 = viewpoints2[["gid", elevationFieldName]].groupby(
                by="gid").agg("mean").to_dict("index")
            viewpoints2 = {
                key: viewpoints2[key][elevationFieldName] for key in viewpoints2.keys()}
            # FROM: ...apply(lambda row: ... if ... else None)
            # TO: ...apply(lambda row: ... if ... else h0)
            # Debug 30.04.2024
            smapRaysField["__HAUTEUR_VP__"] = smapRaysField.apply(
                lambda row: viewpoints2[row.gid] if row.gid in viewpoints2 else h0,
                axis=1)
        # smapRaysField.to_csv("/tmp/4.csv") # DEBUG

        # Debug 06.05.2024
        # smapRaysField = smapRaysField.dissolve(
        #     by="__VPT_ID__", as_index=False, aggfunc={
        #         "__RAY_ID__": list, "gid": "first", "viewpoint": "first", "__HAUTEUR_VP__": "first"})
        aggfunc = dict({"__RAY_ID__": list})
        for f in smapRaysField.columns:
            if f not in ["__VPT_ID__", "__RAY_ID__", "geometry"]:
                aggfunc[f] = "first"
        smapRaysField = smapRaysField.dissolve(
            by="__VPT_ID__", as_index=False, aggfunc=aggfunc)
        smapRaysField.drop(columns=["__VPT_ID__"], inplace=True)
        # smapRaysField.to_csv("/tmp/5.csv") # DEBUG

        if (0 < len(smapRaysField)):
            smapRaysField["__RAY_LEN__"] = smapRaysField.apply(
                lambda row: RayCasting25DLib.__from25DRaysToRayLengths(nRays, row.__RAY_ID__, row.geometry), axis=1)
            smapRaysField["__RAY_ALT__"] = smapRaysField.apply(
                lambda row: RayCasting25DLib.__from25DRaysToRayAlts(nRays, row.__RAY_ID__, row.geometry, row.__HAUTEUR_VP__), axis=1)
            smapRaysField["__RAY_DELTA_ALT__"] = smapRaysField.apply(
                lambda row: RayCasting25DLib.__from25DRaysToRayDeltaAlts(nRays, row.__RAY_ID__, row.geometry, row.__HAUTEUR_VP__), axis=1)

            if withIndices:
                smapRaysField = RayCasting25DLib.__addIndicesToSkyMapRaysField25D(
                    smapRaysField)
        # smapRaysField.to_csv("/tmp/6.csv") # DEBUG

        return smapRaysField

    @staticmethod
    def __addIndicesToSkyMapRaysField25D(smapRaysField):
        smapRaysField["w_mean"] = smapRaysField.__RAY_LEN__.apply(
            lambda raylens: float(raylens.mean()))
        smapRaysField["w_std"] = smapRaysField.__RAY_LEN__.apply(
            lambda raylens: float(raylens.std()))
        smapRaysField["h_mean"] = smapRaysField.__RAY_ALT__.apply(
            lambda heights: float(heights.mean()))

        smapRaysField["h_over_w"] = smapRaysField.apply(
            lambda row: row.__RAY_DELTA_ALT__.mean() / (2 * row.w_mean), axis=1)
        smapRaysField["svf"] = smapRaysField.apply(
            lambda row: SVFLib.svf2018(row.__RAY_DELTA_ALT__, row.__RAY_LEN__), axis=1)

        return smapRaysField



'''
Created on 27 sep. 2024

@author: tleduc

Copyright 2020-2024 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from pymap3d import Ellipsoid


class SatelliteLib():
    '''
    classdocs
    '''
    CONSTELLATIONS = {"E": "Galileo", "G": "GPS", "R": "GLONASS"}
    NSAT = 94
    WGS84 = Ellipsoid.from_name("wgs84")

    @staticmethod
    def constellation(satName):
        firstLetter = satName[0]
        return SatelliteLib.CONSTELLATIONS[firstLetter]

    @staticmethod
    def get_satellite_name(i):
        if (0 <= i <= 31):
            return f"G{i+1:02d}"  # GPS
        if (32 <= i <= 57):
            return f"R{i-32+1:02d}"  # GLONASS
        if (58 <= i <= 93):
            return f"E{i-58+1:02d}"  # GALILEO
        raise Exception("Unreachable instruction!")



'''
Created on 15 jan. 2024

@author: tleduc

Copyright 2020-2024 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
import warnings
from t4gpd.commons.GeoProcess import GeoProcess
from t4gpd.commons.WarnUtils import WarnUtils


class MeteoFrancePredictor(GeoProcess):
    '''
    classdocs

    Données climatologiques de base - horaires
    https://meteo.data.gouv.fr/datasets/donnees-climatologiques-de-base-horaires
    '''

    def __init__(self, meteo, groupby=["month", "hour"]):
        '''
        Constructor
        '''
        warnings.formatwarning = WarnUtils.format_Warning_alt
        warnings.warn("Deprecated class: Use MeteoFranceAggregator instead")
        self.meteo = meteo
        self.groupby = groupby

    @staticmethod
    def __agg(gdf, groupby):
        num_fields = list([f for f in gdf.select_dtypes(
            include="number") if ("NUM_POSTE" != f)])
        aggfunc = {}
        for fieldname in gdf.columns:
            if ("geometry" != fieldname):
                if fieldname in num_fields:
                    aggfunc[fieldname] = "mean"
                else:
                    aggfunc[fieldname] = "first"
        gdf = gdf.dissolve(by=groupby, aggfunc=aggfunc)
        gdf.drop(columns=["timestamp", "AAAAMMJJHH"], inplace=True)
        return gdf

    def run(self):
        return MeteoFrancePredictor.__agg(self.meteo, self.groupby)


"""
from t4gpd.wrf.MeteoFranceReader import MeteoFranceReader

PERIODS = [
    # (1970, 1979),
    (1980, 1989),
    (1990, 1999),
    (2000, 2009),
    (2010, 2019),
    (2020, 2022),
    (2023, 2024),
]

rootdir = "/home/tleduc/data/meteofrance/data"
ifiles = [
    f"{rootdir}/HOR_departement_44_periode_{year0}-{year1}.csv"
    for year0, year1 in PERIODS
]
gdf = MeteoFranceReader(ifiles, station="NANTES-BOUGUENAIS").execute()

import matplotlib.pyplot as plt

for year0, year1 in PERIODS:
    gdf2 = gdf[ (year0 <= gdf.year) & (gdf.year <= year1) ]
    gdf3 = MeteoFrancePredictor(gdf2).execute()

    fig, axes = plt.subplots(nrows=3, ncols=4, figsize=(1.5 * 8.26, 1 * 8.26))
    fig.suptitle(f"Period {year0} - {year1}", fontsize=24)
    for month in range(12):
        nr, nc = month // 4, month % 4
        _gdf = gdf3[(gdf3.month == (1+month))]
        ax = axes[nr, nc]
        ax.set_title(f"Month {month+1}")
        ax.plot(_gdf.hour, _gdf.DNI, color="blue", label="DNI")
        ax.plot(_gdf.hour, _gdf.DHI, color="red", label="DHI")
        ax.plot(_gdf.hour, _gdf.GHI, color="green", label="GHI")
        ax.legend()
    fig.tight_layout()
    plt.savefig(f"DNI_DHI_GHI_{year0}_{year1}.pdf", bbox_inches="tight")
    # plt.show()
    plt.close(fig)
"""



'''
Created on 12 jan. 2024

@author: tleduc

Copyright 2020-2024 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from datetime import timedelta, timezone
from geopandas import GeoDataFrame
from multiprocessing import Pool
from numpy import cos, pi
from pandas import concat, read_csv, to_datetime
from pvlib.irradiance import dirint
from shapely import Point
from t4gpd.commons.ArrayCoding import ArrayCoding
from t4gpd.commons.GeoProcess import GeoProcess
from t4gpd.commons.sun.SunLib import SunLib


class MeteoFranceReader(GeoProcess):
    '''
    classdocs

    Données climatologiques de base - horaires
    https://meteo.data.gouv.fr/datasets/donnees-climatologiques-de-base-horaires
    '''

    def __init__(self, ifiles, station="NANTES-BOUGUENAIS", tzinfo=timezone.utc,
                 ocrs="epsg:2154", model="pysolar", encode=False):
        '''
        Constructor
        '''
        self.ifiles = ifiles
        self.station = station
        # DEBUG DU 17.01.2024 :
        # Les releves des stations MeteoFrance sont en temps UTC
        self.tzinfo = tzinfo
        self.ocrs = ocrs
        self.model = model
        self.encode = encode

    @staticmethod
    def __read(ifile, station, tzinfo):
        df = read_csv(ifile, decimal=".", encoding="utf8", sep=";")
        if station is not None:
            df = df[df.NOM_USUEL == station]
        df["timestamp"] = df.AAAAMMJJHH.apply(
            lambda v: to_datetime(str(v), format="%Y%m%d%H", utc=False).
            replace(tzinfo=tzinfo))
        cols = {}
        for fieldname in [" FXI", " DHUMEC", " N", " SOL", " T", " TUBENEIGE", " UV"]:
            if fieldname in df:
                cols[fieldname] = fieldname.replace(" ", "")
        df.rename(columns=cols, inplace=True)
        df.reset_index(drop=True, inplace=True)
        # REMOVE EMPTY COLUMNS
        # df.dropna(axis=1, inplace=True, how="all")
        return df

    @staticmethod
    def __enrich(df, ocrs, model, encode):
        # DEBUG DU 17.01.2024 :
        # LE CHAMP GLO2 EST SYSTEMATIQUEMENT RENSEIGNE... CONTRAIREMENT
        # AU CHAMP GLO (QUI NE L'EST QUE DEPUIS LES ANNEES 2000)
        # GLO2 = rayonnement global horaire en heure TSV (en J/cm2)
        # TSV = Temps Solaire Vrai (Apparent solar time or true solar time)
        longitude = df.at[0, "LON"]
        df.timestamp = df.timestamp.apply(
            lambda dt: dt + timedelta(hours=(longitude*12/180)))

        df["year"] = df.timestamp.apply(lambda dt: dt.year)
        df["month"] = df.timestamp.apply(lambda dt: dt.month)
        df["day"] = df.timestamp.apply(lambda dt: dt.day)
        df["hour"] = df.timestamp.apply(lambda dt: dt.hour)
        df["day_of_year"] = df.timestamp.apply(lambda dt: dt.day_of_year)
        df["weekday"] = df.timestamp.apply(lambda dt: dt.weekday())

        df["geometry"] = df.apply(
            lambda row: Point([row.LON, row.LAT]), axis=1)
        gdf = GeoDataFrame(df, crs="epsg:4326").to_crs(ocrs)

        if model is not None:
            # Global Horizontal Irradiance (GHI) is the total amount of
            # shortwave radiation received from above by a surface
            # horizontal to the ground.

            # CONVERT GLO [J/cm2] -> GHI [W/m2]
            # rayonnement global horaire en heure UTC (en J/cm2)
            # gdf["GHI"] = (1e4/3600) * gdf.GLO

            # CONVERT GLO2 [J/cm2] -> GHI [W/m2]
            # rayonnement global horaire en heure TSV (en J/cm2)
            gdf["GHI"] = (1e4/3600) * gdf.GLO2

            # DETERMINE ZENITH ANGLES
            sunModel = SunLib(gdf, model=model)
            gdf["solar_alti"], gdf["solar_azim"] = gdf.apply(
                lambda row: sunModel.getSolarAnglesInDegrees(row.timestamp),
                axis=1, result_type="expand").to_numpy().transpose()

            gdf["sun_beam_dir"] = gdf.timestamp.apply(
                lambda dt: sunModel.getRadiationDirection(dt))
            if encode:
                gdf.sun_beam_dir = gdf.sun_beam_dir.apply(
                    lambda t: ArrayCoding.encode(t))

            gdf["solar_zenith"] = gdf.solar_alti.apply(
                lambda alti: 90-alti if (0 < alti) else 0)

            # Direct Normal Irradiance (DNI) is the amount of solar radiation
            # received per unit area by a surface that is always held perpendicular
            # (or normal) to the rays that come in a straight line from the
            # direction of the sun at its current position in the sky.

            # DETERMINE DNI FROM GHI USING THE DIRINT MODIF. OF THE DISC MODEL
            gdf.set_index("timestamp", drop=False, inplace=True)
            gdf["DNI"] = dirint(gdf.GHI, gdf.solar_zenith, gdf.index,
                                pressure=100*gdf.PSTAT,
                                use_delta_kt_prime=True,
                                temp_dew=gdf.TD,
                                min_cos_zenith=0.065,
                                max_zenith=87)

            gdf["cos_solar_zenith"] = (pi/180) * gdf.solar_zenith
            gdf.cos_solar_zenith = cos(gdf.cos_solar_zenith)

            # Diffuse Horizontal Irradiance (DHI) is the amount of radiation
            # received per unit area by a surface (not subject to any shade or
            # shadow) that does not arrive on a direct path from the sun, but
            # has been scattered by molecules and particles in the atmosphere
            # and comes equally from all directions

            # GHI = DNI x cos(theta) + DHI
            gdf["DHI"] = gdf.apply(lambda row: row.GHI -
                                   row.cos_solar_zenith*row.DNI, axis=1)

            gdf.drop(columns=["solar_zenith",
                     "cos_solar_zenith"], inplace=True)
            gdf.reset_index(drop=True, inplace=True)

        return gdf

    @staticmethod
    def _process(params):
        ifile, station, tzinfo, ocrs, model, encode = params
        df = MeteoFranceReader.__read(ifile, station, tzinfo)
        gdf = MeteoFranceReader.__enrich(df, ocrs, model, encode)
        return gdf

    def run(self):
        # PREPROCESS
        listOfParams = [
            (ifile, self.station, self.tzinfo, self.ocrs, self.model, self.encode)
            for ifile in self.ifiles
        ]

        # PROCESS
        pool = Pool()
        gdfs = pool.map(MeteoFranceReader._process, listOfParams)
        pool.close()

        # POSTPROCESS
        gdf = concat(gdfs, ignore_index=True)
        gdf.sort_values(by="timestamp", ascending=True, ignore_index=True,
                        inplace=True)
        return gdf


"""
PERIODS = [ (1900 + i, 1909 + i) for i in range(40, 120, 10) ] + [ (2020, 2022), (2023, 2024) ]
PERIODS = [ (2020, 2022), (2023, 2024) ]

rootdir = "/home/tleduc/data/meteofrance/data"
ifiles = [
    f"{rootdir}/HOR_departement_44_periode_{year0}-{year1}.csv"
    for year0, year1 in PERIODS
]
gdf = MeteoFranceReader(ifiles, station="NANTES-BOUGUENAIS").execute()

_gdf = gdf.loc[gdf[(gdf.day_of_year.isin([5, 6])) & (gdf.year == 2020)].index]

import matplotlib.pyplot as plt
fig, ax = plt.subplots(figsize=(1 * 8.26, 1 * 8.26))
ax.plot(_gdf.timestamp, _gdf.DNI, color="blue", label="DNI")
ax.plot(_gdf.timestamp, _gdf.DHI, color="red", label="DHI")
ax.plot(_gdf.timestamp, _gdf.GHI, color="green", label="GHI")
ax.legend()
plt.savefig("DNI_DHI_GHI.pdf", bbox_inches="tight")
plt.show()
plt.close(fig)
"""



'''
Created on 15 oct. 2024

@author: tleduc

Copyright 2020-2024 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from numpy import ndarray
from t4gpd.commons.IllegalArgumentTypeException import IllegalArgumentTypeException


class IntervalLib(object):
    '''
    classdocs
    '''

    @staticmethod
    def merge(intervals):
        if not isinstance(intervals, (list, ndarray, tuple)):
            raise IllegalArgumentTypeException(
                intervals, "list, numpy ndarray, or tuple of intervals")
        # Sort by start time
        _intervals = sorted(intervals, key=lambda x: x[0])

        result = []
        for interval in _intervals:
            if not result or result[-1][1] < interval[0]:  # No overlap
                result.append(interval)
            else:  # Overlap, merge intervals
                result[-1][1] = max(result[-1][1], interval[1])

        return result
